! ========================================================================
! GsExternalSession.gs
!
! Copyright (C) by GemTalk Systems 1991-2019.  All Rights Reserved
! ========================================================================


! Provide placeholders for references to classes that cannot be generated yet.
! (The classes will be generated during the "bomlastconv.gs" phase.
doit
	Globals
		at: #GciLibrary
			ifAbsentPut: [nil];
		at: #GciErrSType
			ifAbsentPut: [nil]. 
	^true
%

expectvalue %String
doit
(Object subclass: 'GsExternalSession'
	instVarNames: #( parameters gciSessionId stoneSessionId stoneSessionSerial gemProcessId gciErrSType nbResult logger)
	classVars: #()
	classInstVars: #()
	poolDictionaries: #()
	inDictionary: Globals
	options: #()) definition
%

! ------------------- Class comment for GsExternalSession
doit
GsExternalSession comment: 
'GsExternalSession provides a means of running Smalltalk expressions in a different Gem,
possibly on another server or against a different Stone. The expression to execute can
be provided as a String (via #executeString: and #forkString:) or as a Block (via #executeBlock:
and #forkBlock:).

The #executeBlock: protocol allows the use of all the usual Smalltalk tools for working
with Smalltalk code, but does not support arguments to the Block and also requires that
the code compiles in the local image.

The #executeString: protocol supports the use of any String, but being a String, it is
opaque to the operation of most Smalltalk tools, such as the Refactoring Browser, for example.

The "execute" methods are synchronous; the "fork" methods are asynchronous and require checking
for results sometime later. When forking execution, use the #isResultAvailable, #waitForResult (and
variations), and #lastResult methods to get the results of the remote execution.

The Gems launched by the external sessions are not managed by the system, so ensure sessions are 
logged out to prevent lingering sessions consuming system resources unnecessarily.


Examples:

UserGlobals at: #Sample put: GsExternalSession new. "for testing; instances should not be persisted"
Sample 
	stoneNRS: GsNetworkResourceString defaultStoneNRSFromCurrent;
	gemNRS: GsNetworkResourceString defaultGemNRSFromCurrent;
	username: ''DataCurator'';
	password: ''swordfish'';
	yourself.
Sample login.
Sample executeString: ''System stoneName''.
Sample executeBlock: [2 + 5].
Sample executeBlock: [nil].
Sample executeString: ''true''.
Sample logout.
'.
true
%

doit
GsExternalSession category: 'External Sessions'.
true
%

! Remove existing behavior from GsExternalSession
doit
GsExternalSession removeAllMethods.
GsExternalSession class removeAllMethods.
true
%
! ------------------- Class methods for GsExternalSession
category: 'GciLibrary'
classmethod: GsExternalSession
gciErrSTypeClass

	^GciErrSType.
%
category: 'GciLibrary'
classmethod: GsExternalSession
gciLibrary

	^SessionTemps current
		at: #'gciLibrary'
		ifAbsentPut: [
			GciLibrary new
				GciInit;
				yourself.
		].
%
category: 'Instance Creation'
classmethod: GsExternalSession
gemNRS: gemNRS stoneNRS: stoneNRS username: aUsername password: aPassword

	^self new
		gemNRS: gemNRS;
		stoneNRS: stoneNRS;
		username: aUsername;
		password: aPassword;
		yourself
%
category: 'Instance Creation'
classmethod: GsExternalSession
gemNRS: gemNRS stoneNRS: stoneNRS username: gsUsername password: gsPassword hostUsername: hostUsername hostPassword: hostPassword

	^self new
		gemNRS: gemNRS;
		stoneNRS: stoneNRS;
		username: gsUsername;
		password: gsPassword;
		hostUsername: hostUsername;
		hostPassword: hostPassword;
		yourself
%
category: 'Instance Creation'
classmethod: GsExternalSession
new
	"Note that this method is meant to be replaced when
	 the GsLegacyExternalSession class is loaded."
	 
	^super new
		initialize;
		yourself
%
category: 'Instance Creation'
classmethod: GsExternalSession
newDefault
  "This creates an external session that is set to the user, host,
   and stone of the current gem with the password 'swordfish'. This
   is enough to use in an environment where NetLDI uses a captive
   account, but is not sufficient for a more complex environment or
   one where the password has been changed. In that case you use 
   other Instance Creation methods." 

	^self new
		initializeDefaultResources;
		yourself
%

! ------------------- Instance methods for GsExternalSession
category: 'Public'
method: GsExternalSession
abort
	"Abort the current transaction in the external Gem."

	self _errorIfCallInProgress.
	self _gciLibrary GciNbAbort.
	self waitForResult.
	nbResult ~~ 0 ifTrue: [self error: 'Unexpected result!'].
%
category: 'Private'
method: GsExternalSession
clearStackFor: anError

	| contextOop |
	(contextOop := anError context) == nil asOop ifTrue: [^self].
	self _errorIfCallInProgress.
	self _gciLibrary GciClearStack_: contextOop.
	self _signalIfError.
%
category: 'Public'
method: GsExternalSession
commit
	"Commit the current transaction in the external Gem."

	| result |
	self _errorIfCallInProgress.
	result := self _gciLibrary GciCommit.
	self _signalIfError.
	^result == 1.
%
category: 'Public'
method: GsExternalSession
continue: contextOop
	"Continue execution in the external Gem following an exception."

	^self 
		continue: contextOop 
		replacingTopOfStackWithOop: 1	"src/gcioop.ht:#define OOP_ILLEGAL         ((OopType)0x01)" 
%
category: 'Private'
method: GsExternalSession
continue: contextOop replacingTopOfStackWithOop: valueOop
	"Continue execution in the external Gem following an exception,
	 replacing the top of the stack with the specified object.
	 It is an error to specify an oop not visible to the remote Gem."

	self _errorIfCallInProgress.
	self _gciLibrary
		GciNbContinueWith_: contextOop 
		_: valueOop
		_: 0
		_: nil.
	^self 
		waitForResult;
		lastResult.
%
category: 'Public'
method: GsExternalSession
continue: contextOop with: anObject
	"Continue execution in the external Gem following an exception,
	 replacing the top of the stack with the specified object.
	 It is an error if anObject is not visible to the remote Gem."

	^self 
		continue: contextOop 
		replacingTopOfStackWithOop: anObject asOop
%
category: 'Public'
method: GsExternalSession
executeBlock: aBlock
	"Execute the code in the Block argument in the external Gem
	 and answer the result. The best values to return are specials.
	 Strings and other byte objects are copied to the local Gem.
	 All other responses are returned as a Array containing a single OOP."

	| string |
	string := aBlock _sourceString.
	string := string copyFrom:(string indexOf: $[) + 1 to: string size -1.
	^self executeString: string.
%
category: 'Public'
method: GsExternalSession
executeBlock: aBlock with: aValue
	"Execute the code in the Block argument in the external Gem, passing
	 in the specified value, and answer the result. The value passed to the
	 Block must be one whose printString allows the correct object state to
	 be recreated (such as numbers and strings, for example).
	 The best values to return are specials.
	 Strings and other byte objects are copied to the local Gem.
	 All other responses are returned as a Array containing a single OOP."

	^self executeBlock: aBlock withArguments: {aValue}.
%
category: 'Public'
method: GsExternalSession
executeBlock: aBlock with: aValue with: anotherValue
	"Execute the code in the Block argument in the external Gem, passing
	 in the specified values, and answer the result. The values passed to the
	 Block must be ones whose printString allows the correct object state to
	 be recreated (such as numbers and strings, for example).
	 The best values to return are specials.
	 Strings and other byte objects are copied to the local Gem.
	 All other responses are returned as a Array containing a single OOP."

	^self executeBlock: aBlock withArguments: {aValue. anotherValue}.
%
category: 'Public'
method: GsExternalSession
executeBlock: aBlock withArguments: someValues
	"Execute the code in the Block argument in the external Gem, passing
	 in the specified values, and answer the result. The values passed to the
	 Block must be ones whose printString allows the correct object state to
	 be recreated (such as numbers and strings, for example).
	 The best values to return are specials.
	 Strings and other byte objects are copied to the local Gem.
	 All other responses are returned as a Array containing a single OOP."

	| string stream |
	aBlock numArgs = someValues size ifFalse: [self error: 'Wrong number of arguments'].
	stream := ReadWriteStreamPortable on: String new.
	stream nextPutAll: aBlock _sourceString.
	stream nextPutAll: ' valueWithArguments: {'.
	someValues doWithIndex: [:each :index |
		index > 1 ifTrue: [stream nextPutAll: '. '].
		each printOn: stream.
	].
	stream nextPutAll: '}'.
	string := stream reset; skipTo: $^; upToEnd.
	^self executeString: string.
%
category: 'Public'
method: GsExternalSession
executeString: aString
	"Execute the string expression in the external Gem and answer the result.
	 The best values to return are specials.
	 Strings and other byte objects are copied to the local Gem.
	 All other responses are returned as a Array containing a single OOP."

	^self
		forkString: aString;
		waitForResult;
		lastResult.
%
category: 'Public'
method: GsExternalSession
forceLogout

	stoneSessionId ifNil: [^self].
	[self hardBreak] onException: Error do: [:ex | ].
	self _logout.
%
category: 'Public'
method: GsExternalSession
forkBlock: aBlock
	"Execute the code in the Block argument in the external Gem
	 and do not wait for a result. At some later point, you would
	 check for a result. Otherwise you cannot issue another call, as
	 the current call would remain in progress. Refer to #executeString:
	 for an example of the complete send, wait, response sequence."

	| string |
	string := (aBlock _sourceString subStrings: $^) last.
	string := string copyFrom:(string indexOf: $[) + 1 to: string size -1.
	self forkString: string.
%
category: 'Public'
method: GsExternalSession
forkBlock: aBlock with: aValue
	"Execute the code in the Block argument in the external Gem, passing
	 in the specified value, and do not wait for a result. The value passed 
	 to the Block must be one whose printString allows the correct object 
	 state to be recreated (such as numbers and strings, for example).
	 At some later point, you would check for a result. Otherwise you cannot
	 issue another call, as the current call would remain in progress.  
	 Refer to #executeString: for an example of the complete send, wait, response sequence."

	self forkBlock: aBlock withArguments: {aValue}.
%
category: 'Public'
method: GsExternalSession
forkBlock: aBlock with: aValue with: anotherValue
	"Execute the code in the Block argument in the external Gem, passing
	 in the specified values, and do not wait for a result. The values passed 
	 to the Block must be ones whose printString allows the correct object 
	 state to be recreated (such as numbers and strings, for example).
	 At some later point, you would check for a result. Otherwise you cannot
	 issue another call, as the current call would remain in progress.  
	 Refer to #executeString: for an example of the complete send, wait, response sequence."

	self forkBlock: aBlock withArguments: {aValue. anotherValue}.
%
category: 'Public'
method: GsExternalSession
forkBlock: aBlock withArguments: someValues
	"Execute the code in the Block argument in the external Gem, passing
	 in the specified values, and do not wait for a result. The values passed 
	 to the Block must be ones whose printString allows the correct object 
	 state to be recreated (such as numbers and strings, for example).
	 At some later point, you would check for a result. Otherwise you cannot
	 issue another call, as the current call would remain in progress.  
	 Refer to #executeString: for an example of the complete send, wait, response sequence."

	| string stream |
	aBlock numArgs = someValues size ifFalse: [self error: 'Wrong number of arguments'].
	stream := ReadWriteStreamPortable on: String new.
	stream nextPutAll: aBlock _sourceString.
	stream nextPutAll: ' valueWithArguments: {'.
	someValues doWithIndex: [:each :index |
		index > 1 ifTrue: [stream nextPutAll: '. '].
		each printOn: stream.
	].
	stream nextPutAll: '}'.
	string := stream reset; skipTo: $^; upToEnd.
	self forkString: string.
%
category: 'Public'
method: GsExternalSession
forkString: aString
	"Execute the string expression in the external Gem and do not wait for 
	 a result.  At some later point, you would check for a result. Otherwise you cannot
	 issue another call, as the current call would remain in progress.  
	 Refer to #executeString: for an example of the complete send, wait, response sequence."

	self _errorIfCallInProgress.
	self _gciLibrary 
		GciNbExecuteStr_: aString 
		_: nil asOop.
%
! fix 47868
category: 'Parameters'
method: GsExternalSession
gemNRS: anNRS
	"Set the GemService parameters for the logon to the value
	 of anNRS, which may be a String or a GsNetworkResourceString instance."

	parameters gemService: anNRS asString .
  (anNRS isKindOf: GsNetworkResourceString) ifTrue:[	
    self dynamicInstVarAt: #gemHost put: anNRS node .
  ].
%
category: 'Accessors'
method: GsExternalSession
gemProcessId

  ^gemProcessId.
%
category: 'Public'
method: GsExternalSession
hardBreak
	"Interrupt the external Gem and abort the current transaction."

	self _setSessionId.
	self _gciLibrary GciHardBreak.
	self _signalIfError.
%
category: 'Parameters'
method: GsExternalSession
hostPassword: aString

	parameters hostPassword: aString copy.
%
category: 'Parameters'
method: GsExternalSession
hostUsername: aString

	parameters hostUsername: aString copy.
%
category: 'Private'
method: GsExternalSession
initialize

	gciErrSType := self class gciErrSTypeClass new.
	parameters := GemStoneParameters new.
	self loggingToServer.
%
category: 'Private'
method: GsExternalSession
initializeDefaultResources

	self
		gemNRS: GsNetworkResourceString defaultGemNRSFromCurrent;
		stoneNRS: GsNetworkResourceString defaultStoneNRSFromCurrent;
		username: System myUserProfile userId;
		password: 'swordfish';
		yourself.
%
category: 'Public'
method: GsExternalSession
isCallInProgress
	"Answer whether there is currently a call in progress to
	 the external Gem.
	 
	 The following calls are OK during a nonblocking call:
		GciCallInProgress
		GciErr
		GciGetSessionId
		GciHardBreak
		GciNbEnd
		GciSetSessionId
		GciShutdown
		GciSoftBreak"

	| result |
	self isLoggedIn ifFalse:[ ^ false "not logged in"].
	self _setSessionId.
	result := self _gciLibrary GciCallInProgress.
	self _signalIfError.
	^result == 1.
%
category: 'Private'
method: GsExternalSession
_isOnMyStone
	^ self dynamicInstVarAt: #_isOnMyStone
%
category: 'Public'
method: GsExternalSession
isRemoteServerBigEndian

	^self _gciLibrary GciServerIsBigEndian ~~ 0
%
category: 'Public'
method: GsExternalSession
isResultAvailable
	"Check whether the current call in progress has finished
	 and save the result if it has. Most operations cannot be
	 started while another is in progress. You must call this
	 method and receive a true result before starting another
	 remote operation."
	 
	| cByteArray result |
	nbResult := nil.
	self isCallInProgress ifFalse: [self error: 'no call in progress'].
	cByteArray := CByteArray gcMalloc: 8.
	result := self _gciLibrary GciNbEnd_: cByteArray.
	result < 2 ifTrue: [^false].
	self _signalIfError.
	nbResult := (cByteArray pointerAt: 0 resultClass: CByteArray) uint64At: 0.
	^true.
%
category: 'Public'
method: GsExternalSession
lastResult
  ^ nbResult ifNotNil:[:r | 
     "r is an Integer, the value of a remote OopType"
     self resolveResult: r
  ]
%
category: 'Logging'
method: GsExternalSession
log: aString

	logger value: aString
%
category: 'Logging'
method: GsExternalSession
logger: aOneArgBlock
	"Use the specified one-argument Block for logging messages.
	 The argument to the Block is the message to log."

	logger := aOneArgBlock
%
category: 'Logging'
method: GsExternalSession
loggingToClient

	self logger: [:message | GsFile gciLogClient: message]
%
category: 'Logging'
method: GsExternalSession
loggingToServer

	self logger: [:message | GsFile gciLogServer: message]
%
! fixed 47663
category: 'Public'
method: GsExternalSession
login

	| result onMyStn |
	stoneSessionId ifNotNil: [
		ImproperOperation signal: 'Stone session ' , stoneSessionId printString , 
			' already associated with this GsExternalSession!'.
	].
	self _gciLibrary
		GciSetNetEx_: parameters gemStoneName
		_: parameters hostUsername
		_: parameters hostPassword
		_: parameters gemService
		_: parameters passwordIsEncryptedAsIntegerBoolean.	"1 or 0: GCI_LOGIN_PW_ENCRYPTED"
	self _signalIfError.
	result := self _gciLibrary 
		GciLoginEx_: parameters username
		_: parameters password
		_: parameters loginFlags
		_: 0. "haltOnErrNum"
	self _signalIfError.
	0 == result ifTrue: [
		self error: 'Login failed for unknown reason!'.
	].
	gciSessionId := self _gciLibrary GciGetSessionId.
	stoneSessionId := Object _objectForOop: (self _gciLibrary GciPerform_: System asOop _: 'session' _: nil _: 0).
	self _signalIfError.
  onMyStn :=  System stoneStartupId asString 
              = (self executeString:'System stoneStartupId asString') .
  self dynamicInstVarAt: #_isOnMyStone put: onMyStn .
	onMyStn ifTrue: [
		stoneSessionSerial := GsSession serialOfSession: stoneSessionId.
		gemProcessId := (System descriptionOfSession: stoneSessionId) at: 2.
	] ifFalse: [
		stoneSessionSerial := self executeString: 'GsSession currentSession serialNumber'.
		gemProcessId := self executeString: 'System gemVersionAt: #''processId'''.
	].
	self log: 'GsExternalSession login: ' , self _describe.
%
category: 'Public'
method: GsExternalSession
logout
	stoneSessionId ifNil: [^self].
	self isCallInProgress ifTrue: [
		[
			self waitForResult.
		] onException: Error do: [:ex | 
			ex return.
		].
	].
	self _logout.
%

category: 'Private'
method: GsExternalSession
nbLogout
  "Private. 
   Should be followed by a send of _waitForLogout ."

	stoneSessionId ifNil: [^self].
	self isCallInProgress ifTrue: [
		[
			self waitForResult.
		] onException: Error do: [:ex | 
			ex return.
		].
	].
	self _nbLogout .
%

category: 'Parameters'
method: GsExternalSession
password: aString

	parameters password: aString copy.
%
category: 'Public'
Set compile_env: 0
method: GsExternalSession
printOn: aStream

	aStream
		nextPutAll: 'a';
		nextPutAll: self class name;
		nextPutAll: '(';
		nextPutAll: stoneSessionId printString;
		nextPutAll: '/';
		nextPutAll: stoneSessionSerial printString;
		nextPutAll: ')';
		yourself.
%
category: 'Public'
method: GsExternalSession
send: selector to: anOop
	"Answer the result of having the specified remote object
	 sent the message with the specified selector.
	 The best values to return are specials.
	 Strings and other byte objects are copied to the local Gem.
	 All other responses are returned as a Array containing a single OOP."

	^self send: selector to: anOop withArguments: nil
%
category: 'Public'
method: GsExternalSession
send: selector to: anOop withArguments: someValues
	"Answer the result of having the specified remote object
	 sent the message with the specified selector and arguments.
	 Argument values are passed by OOP; beware of inconsistent views
	 between the local Gem and the external Gem.
	 The best values to return are specials.
	 Strings and other byte objects are copied to the local Gem.
	 All other responses are returned as a Array containing a single OOP."

	| args |
	self _errorIfCallInProgress.
	(someValues == nil or: [someValues isEmpty])
		ifTrue: [args := nil]
		ifFalse: 
			[args := CByteArray gcMalloc: 8 * someValues size.
			someValues doWithIndex: 
					[:each :index |
					| bytes |
					bytes := ByteArray new: 8.
					bytes unsigned64At: 1 put: each asOop.
					"bytes := bytes reverse."
					self isRemoteServerBigEndian ifFalse: [bytes := bytes reverse].
					args
						copyFrom: bytes
						from: 1
						to: 8
						into: (index - 1) * 8]].
	self _gciLibrary
		GciNbPerform_: anOop
		_: selector
		_: args
		_: someValues size.
	^self
		waitForResult;
		lastResult
%
category: 'Accessors'
method: GsExternalSession
sessionId
	"0 => not logged in"

	^gciSessionId.
%

category: 'Accessors'
method: GsExternalSession
isLoggedIn
 ^ gciSessionId ~~ nil and:[ gciSessionId > 0].
%

category: 'Public'
method: GsExternalSession
softBreak
	"Interrupt the external Gem, but permit it to be restarted."

	self _setSessionId.
	self _gciLibrary GciSoftBreak.
	self _signalIfError.
%
category: 'Parameters'
method: GsExternalSession
stoneNRS: anNRS
	"Set the Stone parameters for the logon to the value
	 of anNRS, which may be a String or a GsNetworkResourceString instance."

	parameters gemStoneName: anNRS asString
%
category: 'Accessors'
method: GsExternalSession
stoneSessionId

	^stoneSessionId.
%
category: 'Accessors'
method: GsExternalSession
stoneSessionSerial

	^stoneSessionSerial.
%
category: 'Logging'
method: GsExternalSession
suppressLogging

	self logger: [:message | ]
%
category: 'Parameters'
method: GsExternalSession
username: aString

	parameters username: aString.
%
category: 'Parameters'
method: GsExternalSession
username
  ^ parameters username
%
! fix 47021
category: 'Public'
method: GsExternalSession
resolveResult: anOop
	"Answer the object, or Array containing the OOP, of the result 
	received when the last #isResultAvailable answered true. 
	Specials can be fully resolved in the current session, and are the 
	preferred return type. 
	Results that are byte objects will be copied into a String or ByteArray 
	and that will be returned, but the OOP of the byte object will remain 
	in the remote Gem's export set. 
	For objects of all other types, return a result as an Array containing 
	the object's OOP in the remote gem, which is also recorded in the remote 
	Gem's export set."
	 
	| type |
	self _setSessionId.
	type := self _gciLibrary GciFetchObjImpl_: anOop.
	self _signalIfError.
	type == 3 ifTrue: [^Object _objectForOop: anOop]. "result is a special"
	type == 1 ifTrue: [^self _getBytes: anOop].	  "result is a byte object"
	^ { anOop } .
%
! edited for 46919
category: 'Public'
method: GsExternalSession
resolveResult: anOop toLevel: anInteger
	"Similar to resolveResult:, but this recognizes more classes.
	If the class is not recognized, then return a CByteArray 
	with the OOP"

	| library object oop cByteArray |
	object := self resolveResult: anOop.
	(object _isArray) ifFalse: [^object].	"a special or a byte object"
	library := self _gciLibrary.
	oop := library GciFetchClass_: anOop.
	(oop == Array asOop and: [0 < anInteger]) ifTrue: [
		| size array |
		size := library GciFetchSize__: anOop.
		array := Array new.
		1 to: size do: [:i | 
			oop := library GciFetchOop_: anOop _: i. 
			array add: (self resolveResult: oop toLevel: anInteger - 1).
		].
		^array.
	].
	"Not a recognized object"
	cByteArray := CByteArray gcMalloc: 8.
	cByteArray uint64At: 0 put: anOop.
	^cByteArray.
%
category: 'Public'
method: GsExternalSession
waitForResult
	"Wait as long as it takes for the external Gem to complete
	 the current operation."

	self waitForResultForSeconds: PlusInfinity
%
category: 'Public'
method: GsExternalSession
waitForResultForSeconds: aNumber
	"Wait as long as the specified seconds for the external Gem
	 to complete the current operation."

	self 
		waitForResultForSeconds: aNumber
		otherwise: [self error: 
			'Wait time of ' , aNumber printString , 
			' exceeded for session ' , stoneSessionId printString , '/' , stoneSessionSerial printString ,
			' (PID ' , gemProcessId printString , ')']
%
category: 'Public'
method: GsExternalSession
waitForResultForSeconds: aNumber otherwise: aBlock
	"Wait as long as the specified seconds for the external Gem
	 to complete the current operation. If the operation does not
	 complete within that time, answer the result of evaluating aBlock."

	| start |
	start := System timeGmt.
	[self isResultAvailable] whileFalse: 
			[Delay waitForMilliseconds: 20.
			aNumber < (System timeGmt - start) ifTrue: [^aBlock value]]
%
category: 'Private'
method: GsExternalSession
_describe
  | str |
	(str := 'stone session ID ' copy )
     add: stoneSessionId asString ;
     add: ' gem processId '; add: gemProcessId asString .
  self _gemHost ifNotNil:[ :host |
     str add: ' on host ' , host asString.
  ].
  str add: ' stone serialNumber ' ; add: stoneSessionSerial asString.
  ^ str
%

method: GsExternalSession
_gemHost
  ^ self dynamicInstVarAt: #gemHost
%
category: 'Private'
method: GsExternalSession
_errorIfCallInProgress

	self isCallInProgress ifTrue: [self error: 'call in progress'].
%
category: 'Private'
method: GsExternalSession
_gciLibrary

	^self class gciLibrary
%
category: 'Private'
method: GsExternalSession
_getBytes: anOop

	| size bytes fetchedSize classOop |
	classOop := self _gciLibrary GciFetchClass_: anOop.
	self _signalIfError.
	size := self _gciLibrary GciFetchSize__: anOop.
	self _signalIfError.
	bytes := CByteArray gcMalloc: size.
	fetchedSize := self _gciLibrary
		GciFetchBytes__: anOop
		_: 1
		_: bytes
		_: bytes size.
	self _signalIfError.
	fetchedSize ~~ size ifTrue: [self error: 'Unexpected size!'].
	classOop == String asOop ifTrue: [
		^bytes 
			stringFrom: 0
			to: size - 1.
	].
	classOop == Symbol asOop ifTrue: [
		^ Symbol withAll: (bytes 
			stringFrom: 0
			to: size - 1).
	].
	^bytes 
		byteArrayFrom: 0
		to: size - 1.
%
category: 'Private'
method: GsExternalSession
_logout
	| descr |
	descr := self _describe.
	self isLoggedIn ifTrue:[
		self _nbLogout: descr .
		self _waitForLogout.
		gciSessionId := 0.
	].
	stoneSessionId := nil.
	stoneSessionSerial := nil.
	self log: 'GsExternalSession logout: ' , descr.
%

category: 'Private'
method: GsExternalSession
_nbLogout
  | descr |
  descr := self _describe.
	self log: 'GsExternalSession nbLogout: ' , descr.
  ^ self _nbLogout: descr
%

category: 'Private'
method: GsExternalSession
_nbLogout: descr
  self isLoggedIn ifTrue:[
		self _setSessionId.
		self _signalIfError.
		self _gciLibrary GciLogout.
		[ [ self _signalIfError
			] onException: GciError do:[:ex | 
				 ex originalNumber == 4100 
					 ifTrue:[ "ignore invalid session error from GciError"]
					ifFalse:[ ex pass ]. 
			].
		] onException: Error do: [:ex | | msg |
			(msg := '---(During GsExternalSession logout:') , descr lf .
			msg := msg , '    ', ex description , ')---'.
			self log: msg lf .
		].  
  ].  
%

category: 'Private'
method: GsExternalSession
_setSessionId

  self isLoggedIn ifFalse:[
     ^ Error signal:'invalid sessionId (session not logged in)' 
  ].
  self _gciLibrary GciSetSessionId_: gciSessionId .
  self _signalIfError.
%
category: 'Private'
method: GsExternalSession
_signalIfError

	(self _gciLibrary GciErr_: gciErrSType) == 1 ifFalse: [^self].
	(gciErrSType number  between: 4000 and: 4999) ifTrue: [gciSessionId := 0].
	GciError 
		signal: gciErrSType 
		in: self.
%

! fixed 46165
category: 'Private'
method: GsExternalSession
_waitForLogout
	stoneSessionId ifNil: [^self].
	1 to: 2000 do:[ :j |
		self _isOnMyStone ifTrue: [
			(GsSession serialOfSession: stoneSessionId) = stoneSessionSerial ifFalse:[ 
                           ^self
                        ].
		] ifFalse: [
                   (System _hostProcessExists: gemProcessId) ifFalse: [ 
                      ^ self 
                   ].
		].
		(Delay forMilliseconds: 10) wait. 
	].
        self error: 'session with stone session ID of ' , stoneSessionId printString , 
                    ' gemProcessId = ' , gemProcessId printString, 
                    ' still present 20 seconds after logout'.
%
