!=========================================================================
! Copyright (C) GemTalk Systems 1986-2019.  All Rights Reserved.
!
! Superclass Hierarchy:
!   GsSignalingSocket, GsSocket, IO, Object.
!
!=========================================================================

set class GsSignalingSocket
removeallmethods
removeallclassmethods 

category: 'For Documentation Installation only'
classmethod:
installDocumentation

self comment:
'GsSignalingSocket is a subclass GsSocket that signals Errors,
primarily instances of  SocketError, instead of returning nil when a
socket operation fails.'
%

! ------------------- Class methods 
category: 'Queries'
classmethod:
getServicePortByName: serviceName

"Returns the port number for the given service or signals an Error."

 ^ self getServicePortByName: serviceName withProtocol: nil
%

category: 'Queries'
classmethod:
getServicePortByName: serviceName withProtocol: protocol

"Returns the port number for the given service and protocol.
 Protocol may be nil (interpreted as tcp), 'tcp', or 'udp' .
 Signals an Error if the service is undefined."

 (super getServicePortByName: serviceName withProtocol: protocol)
   ifNotNil:[:r | ^ r ].
 self signalError .
%

category: 'Queries'
classmethod:
getServiceNameByPort: servicePort withProtocol: protocol

"Returns the service name for the given port number and protocol.
 Protocol may be nil (interpreted as tcp), 'tcp', or 'udp' .
 Signals an Error if the service is undefined."

 (super getServiceNameByPort: servicePort withProtocol: protocol)
   ifNotNil:[:r | ^ r ].
 self signalError .
%

category: 'Queries'
classmethod:
getHostNameByAddress: address

"Returns the host name for the given ip address.
 Signals an Error if the host is undefined.

 The address needs to be a String in a form acceptable to 
 either inet_pton(AF_INET6, address, ...)
 or    inet_pton(AF_INET, address, ...) in the host operating  
 system."

 (super getHostNameByAddress: address) ifNotNil:[:r | ^ r ].
 self signalError .
%

category: 'Queries'
classmethod:
getHostAddressByName: aHostName

"aHostName must be a String .
 Returns the ip address for aHostName .

 Signals an Error if aHostName cannot be resolved.

 The resulting ip address is a String in a form acceptable to 
 inet_pton(AF_INET6, aString, ...)

 If getaddrinfo() returns more than one address for aHostName,
 the first element of the list from getaddrinfo() is returned.
"

 (super getHostAddressByName: aHostName) ifNotNil:[:r | ^ r ].
 self signalError .
%


category: 'Queries'
classmethod:
getLocalHostName

"Returns the name of the local host or signals an Error."

 (super getLocalHostName) ifNotNil:[:r | ^ r ].
 self signalError .
%


category: 'Configuration'
method:
raiseExceptionOnError: aBoolean

"Sets the value of raiseExceptionOnError to aBoolean for the receiver.
 Returns the receiver."

 aBoolean ~~ true ifTrue:[ ArgumentError signal:'should not shutoff errors in GsSignalingSocket'].
 super raiseExceptionOnError: aBoolean .
%

category: 'Instance Creation'
classmethod:
new
"Returns a new non-blocking TCP socket 
 or signals an Error if unable to create a new socket.

 On Solaris this creates a socket with family AF_INET6.

 On other platforms this creates a socket with family AF_INET.
 which may be bound or connected to IPv4 addresses or to IPv4-mapped IPv6 addresses."

 | s |
 s := self _basicNew .
 (s _zeroArgPrim: 1) ifNil:[ self signalError ].
 s raiseExceptionOnError: true .
 ^ s
%

classmethod:
newUdp

"Returns a new non-blocking UDP socket 
 or signals an Error if unable to create a new socket.

 On Solaris this creates a socket with family AF_INET6.

 On other platforms this creates a socket with family AF_INET
 which may be bound or connected to IPv4 addresses or to IPv4-mapped IPv6 addresses."

 | s |
 s := self _basicNew .
 (s _zeroArgPrim: 26) ifNil:[ self signalError ].
 s raiseExceptionOnError: true .
 ^ s
%

category: 'Instance Creation'
classmethod:
newIpv6

"Returns a new non-blocking TCP socket 
 or signals an Error if unable to create a new socket.

 This creates a socket with family AF_INET6
 which may be bound or connected to IPv6 addresses.
 On Linux and solaris the result socket may also 
 be bound or connected to IPv4-mapped IPv6 addresses."

 | s |
 s := self _basicNew .
 (s _zeroArgPrim: 34) ifNil:[ self signalError ].
 s raiseExceptionOnError: true .
 ^ s
%

classmethod:
newUdpIpv6

"Returns a new non-blocking UDP socket 
 or signals an Error if unable to create a new socket.

 This creates a socket with family AF_INET6
 which may be bound or connected to IPv6 addresses.
 On Linux and solaris the result socket may also 
 be bound or connected to IPv4-mapped IPv6 addresses."

 | s |
 s := self _basicNew .
 (s _zeroArgPrim: 35) ifNil:[ self signalError ].
 s raiseExceptionOnError: true .
 ^ s
%

classmethod:
fromFileHandle: aSmallInteger
 "Read carefully the documentation in GsSocket >> fromFileHandle: "

 | r |
 r := super fromFileHandle: aSmallInteger .
 r raiseExceptionOnError: true .
 ^ r
%

category: 'Error Reporting'
classmethod: 
signalError
  ^ SocketError signal: (self lastErrorString ifNil:[ 'no details' ]).
%
method: 
signalError
  ^ SocketError signal: (self lastErrorString ifNil:[ 'no details' ]).
%
classmethod: 
signalError: aString
  ^ SocketError signal: 
      (self lastErrorString ifNotNil:[:s | aString , ', ' , s]  ifNil:[ aString ]).
%
method: 
signalError: aString
  ^ SocketError signal: 
     (self lastErrorString ifNotNil:[:s | aString , ', ' , s]  ifNil:[ aString ]).
%

! ------------------- Instance methods 

category: 'Client Operations'
method:
bindTo: portNumber

"Binds the receiver to the specified port number.  Use the makeServer
 methods to do the bind when creating a server socket.
 If portNumber is nil then a random port is selected.
 This method is provided to bind a client socket to a specific port
 before it is connected.  Returns the port number actually
 bound to (should be the same as the argument unless argument is nil), 
 or signals an Error if not successful."

 ^ self bindTo: portNumber toAddress: nil
%

category: 'Client Operations'
method:
bindTo: portNumber toAddress: address

"Binds the receiver to the specified port number and address. 
 Use the makeServer methods to do the bind when creating a server socket.
 This method is provided to bind a client socket to a specific port
 and/or address before doing a listen or connect on the socket.
 If portNumber is nil then a random port is selected.
 If address is nil then any network interface(IN6ADDR_ANY_INIT) is used.
 Returns the port number actually bound to (should be the same as 
 the argument unless argument is nil), 
 or signals an Error if not successful.

 address is a String containing a numeric IP address in one of these forms
   IPv4  dotted-decimal format  d.d.d.d  
   IPv6  hex format   x:x:x:x:x:x:x:x , 
   IPv4-mapped IPv6  ::FFFF:d.d.d.d, 
   where d is an 8 bit decimal number and x is a 16 bit hexadecimal number.
   An IPv6 format may contain at most one :: which is a contigous
   group of zeros.

 Gs64 v3.1:  address = -1, denoting IPv4 INADDR_BROADCAST, no longer supported.
 
"
 (super bindTo: portNumber toAddress: address) ifNotNil:[:r | ^ r ].
 self signalError .
%


! HR 8681, fix 42571
category: 'Client Operations'
method:
connectTo: portNumber on: aHost timeoutMs: timeoutMs

"Connect the receiver to the server socket identified by portNumber and aHost.
aHost may be the name of the host or its numeric address,
or aHost == -1 for <broadcase> , or aHost == nil for IN6ADDR_ANY_INIT .
portNumber maybe either a SmallInteger, or the String name of a service.
timeoutMs is a SmallInteger specifying maximum time to wait for the
connection to complete, or -1 to wait indefinitely.
If aHost is the name of a host, connect is attempted on IPv4 first 
if getaddrinfo returns any IPv4 addresses for aHost, then attempted
on IPv6.
Returns true if the connection succeeded otherwise signals an Error."

(super connectTo: portNumber on: aHost timeoutMs: timeoutMs) ~~ true
  ifTrue:[ self signalError:'connect failed' ].
^ true
%


category: 'Accessing'
method:
option: optionName

"Returns the current value of the specified option.
 Signals an Error if an error occurred."

 (super option: optionName) ifNotNil:[:r | ^ r ]. 
 self signalError .
%

category: 'Socket Operations'
method:
option: optionName put: value

"Sets the value of the specified option.
 Returns the receiver or signals an Error."

 (super option: optionName put: value) ifNotNil:[:r | ^ r ].
 self signalError .
%
   
category: 'Socket Operations'
method:
linger: bool length: timeOut

"Sets up the receiver so that if unsent data is waiting to be transmitted
 at the time the receiver is closed, the current GsProcess will 
 be suspended until either the data is transmitted, or the given timeOut expires.  
 timeOut is in units of seconds.

 Returns the receiver or signals an Error."

 (super linger: bool length: timeOut) ifNotNil:[:r | ^ r ].
 self signalError .
%

category: 'Server Operations'
method:
makeListener: queueLength

"Turns the receiver into a listening socket.
 The queueLength argument specifies the size of the listen backlog queue for
 incoming connections.
 Returns the receiver or signals an Error."

 (super makeListener: queueLength) ifNotNil:[:r | ^ r ].
 self signalError .
%


category: 'Server Operations'
method:
accept

"Accept a client request for a connection on the receiver.  
 Returns a socket created for a new connection, 
 or signals an Error if there was some problem.
 The result is an instance of speciesForAccept , and has non-blocking state
 equal to non-blocking state of the receiver."

 (super accept) ifNotNil:[:r | 
   r raiseExceptionOnError: true .
   ^ r 
 ].
 self signalError .
%

category: 'Server Operations'
method:
acceptTimeoutMs: timeoutMs
 "Returns a socket created for a new connection,
  or signals an Error if there was an error or a timeout.
  The result is an instance of speciesForAccept , and has non-blocking state
  equal to non-blocking state of the receiver."
| res aSocket cnt |
aSocket := self speciesForAccept _basicNew .
res := true.
cnt := 1 .
[ res == true or:[ res == false ]] whileTrue: [
  "primitive will wait until data available if C socket is blocking."
  res := self _twoArgPrim: 4 with: aSocket with: nil .
  (res == false) ifTrue: [ 
    "socket is non-blocking and would have blocked, process scheduler 
     will suspend the current GsProcess"
    res := self readWillNotBlockWithin: timeoutMs .
    res ifNil:[ ^ self signalError ].
    res == false ifTrue:[ ^ self signalError:'accept timed out' ].
  ].
  cnt := cnt + 1 "for debugging"
].
res ifNil:[ ^ self signalError ].
res := aSocket.
^res
%

! peerName inherited
! peerAddress inherited
! peerPort inherited
! port inherited
! address inherited

category: 'Reading'
method:
read: maxBytes into: byteObj startingAt: index

"Reads up to the given number of bytes into the given byte object (for
 example, a String).  The first byte read will go into the position
 indicated by index. 
 Returns the number of bytes read, or 0 if EOF on the receiver,
 or signals an Error .

 maxBytes must be a SmallInteger > 0 and <= 500000000 .

 An error will be raised if index is greater than the size of byteObj.
 Note that index can be equal to the size of byteObj plus 1 in which
 case any bytes read will be appended to byteObj.

 If no data is available for reading
 the current GsProcess is suspended until data arrives.
 The readWillNotBlock or readWillNotBlockWithin: methods may be used to
 determine whether or not data is ready for reading before calling this method."

 (super read: maxBytes into: byteObj startingAt: index) ifNotNil:[:r | ^ r ].
 self signalError .
%

category: 'Reading'
method:
readString: amount untilFalse: aBlock

"Reads bytes from the receiver into a String which it creates and returns.
 Will keep reading until amount bytes are read or 
 the one argument Block  aBlock  returns false.
 Each time aBlock is evaluated it is given the number of bytes read so far.
 aBlock is only evaluated after a partial read.
 Returns the String or signals an Error."

| bytes |
bytes := String new.
self read: amount into: bytes untilFalse: aBlock .
^ bytes.
%

category: 'Reading'
method:
readString: maxBytes

"Reads up to the given number of bytes, returning them in a String whose size is
 between 0 and maxBytes inclusive.  If an error occurs, an Error is signaled.
 EOF is not an error, if EOF occurs on the receiver before any bytes are read,
 a String of size zero is returned.

 maxBytes must be a SmallInteger > 0 .
 If maxBytes is greater than the size of the operating system's buffer for 
 the socket, the size of the result string may be a function of this
 buffer size, even if more data is available from the sender.  Repeated
 invocation of this method may be necessary to obtain all of the data.

 If no data is available for reading
 the current GsProcess is suspended until data arrives.
 The readWillNotBlock or readWillNotBlockWithin: methods may be used to
 determine whether or not data is ready for reading before calling this method."
 
  | bytes |
  bytes := String new.
  self read: maxBytes into: bytes startingAt: 1.
  ^ bytes
%

category: 'Writing'
method:
write: amount from: byteObj startingAt: index

"Write bytes from byteObj to the receiver.
 The first byte written will be from the position indicated by index.
 The number of bytes written is specified by amount.
 It is in an error if amount bytes are not available.
 Returns the number of bytes written, or signals an Error if an error occurs.
  
 If the receiver is not ready for writing
 the current GsProcess is suspended until data arrives.
 The writeWillNotBlock or writeWillNotBlockWithin: methods may be used to
 determine whether or not data is ready for writing before calling this method."
  
 | offset numToWrite waitCnt |
 offset := index.
 numToWrite := amount.
 [ true ] whileTrue:[ | result |
   [
     result := self _write: byteObj startingAt: offset ofSize: numToWrite .
     result == true 
   ] whileTrue . "loop to handle EINTR "
 
   result _isSmallInteger ifTrue:[
     numToWrite  := numToWrite - result .
     (numToWrite <= 0) ifTrue: [ ^ amount ]. "All done"
     waitCnt := nil .
     offset := offset + result .
   ] ifFalse: [
     result ifNil:[ self signalError ].
     waitCnt ifNotNil:[ 
       waitCnt < 3 ifTrue:[
         waitCnt := waitCnt + 1 .
         Delay waitForMilliseconds: 1.
       ] ifFalse:[
         GsFile gciLogServer:'socket EAGAIN fd ', fileDescriptor asString .
         System _printSocketTrace .
         SocketError signal:'EAGAIN from socket write after poll said write-ready'.
       ].
     ] ifNil:[
       waitCnt := 0 .
     ].
     self _waitForWriteReady . 
   ].
 ].
%

category: 'Writing'
method:
nbwrite: amount from: byteObj startingAt: index

"If receiver is a non-blocking socket,
 write at most amount bytes from byteObj to the receiver without blocking.

 If receiver is a blocking socket,
 write at most amount bytes from byteObj to the receiver blocking until
 the socket accepts data, or until the gem process receives SIGTERM.

 The first byte written will be from the position indicated by index.
 The maximum number of bytes to write is specified by amount.
 It is an error if amount bytes are not available.
 Returns the number of bytes written, or signals an Error .
 Returns 0 if non-blocking socket would have blocked."

 | res |
 [
   res := self _write: byteObj startingAt: index ofSize: amount.
   (res == false) ifTrue: [
     "would have blocked"
     ^ 0
   ].
   res == true  "if res==true, socket got EINTR and we retry"
 ] whileTrue .
 res ifNil:[ self signalError ].
 ^ res
%


